#if !defined(VOCAL_ACTION_HXX)
#define VOCAL_ACTION_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const Action_hxx_Version = 
    "$Id: Action.hxx,v 1.2 2001/08/10 04:02:04 icahoon Exp $";


#include "VocalCommon.hxx"
#include "Writer.hxx"


/** Infrastructure common to VOCAL.
 */
namespace Vocal 
{


/** Infrastructure in VOCAL related to "algorithms and the assignment of
 *  responsibilities between objects".<br><br>
 *
 *  See Design Patterns, ISBN 0-201-63361-2, Chapter 5.
 */
namespace Behavioral
{


class Stimulus;


/** An Action object is used to implement an algorithm.<br><br>
 *
 *  Actions can be chained one after another.<br><br>
 *
 *  An example use of an Action can be seen in StateInterface class.
 *  The StateInterface, rather than implementing each response to 
 *  a Stimulus with a method, associates an Action with each Stimulus.
 *  The Action is a more discrete element than a state, and therefore
 *  Actions can be shared among many StateInterfaces or perhaps many
 *  state machines.
 *  <br><br>
 *
 *  @see    Vocal::Behavioral::StateInterface
 *  @see    Vocal::Behavioral::Stimulus
 *  @see    Vocal::IO::Writer
 *  @see    Vocal::ReturnCode
 *
 */
class Action : public Vocal::IO::Writer
{


    public:


    	/** Default constructor. By default, no actions are chained.
	 */
    	Action();
				

	/** Virtual destructor, since you should be extending this class.
	 */
	virtual ~Action();


    	/** The public interface to all Actions are the action() methods.
	 *  This method call's this classes onAction method, and then if 
	 *  another Action is chained to this one, calls the action method 
	 *  on the chained Action. This variation allows an action() to be
	 *  called without a Stimulus.
	 */
    	ReturnCode  	    	action() throw ();


    	/** The public interface to all Actions are the action() methods.
	 *  This method call's this classes onAction method, and then if 
	 *  another Action is chained to this one, calls the action method 
	 *  on the chained Action. This action to be
	 *  called without a Stimulus.
	 */
    	ReturnCode  	    	action(const Stimulus &) throw ();


    	/** Creates an association between this Action and the chained Action,
	 *  such that the chained Action will be done after this one.
	 */
    	virtual void	    	chain(Action &);
	

    	/** Enabled writing this Action to an ostream. By default writes the
	 *  name().
	 */
	virtual ostream &   	writeTo(ostream &) const;
	

	/** The name of the Action. Useful for logging.
	 */
    	virtual const char *	name() const;


    protected:


    	/** Method to be extended by the user of this class. Used by action().
	 */
    	virtual ReturnCode  	onAction(const Stimulus * = 0) throw();


    	// Used for chaining.
	//
    	Action	    	    *	myNextAction;
};


} // namespace Behavioral
} // namespace Vocal


#endif // !defined(VOCAL_ACTION_HXX)
