#if !defined(VOCAL_CONTEXT_HXX)
#define VOCAL_CONTEXT_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const Context_hxx_Version = 
    "$Id: Context.hxx,v 1.2 2001/08/10 04:02:04 icahoon Exp $";


#include "Writer.hxx"


/** Infrastructure common to VOCAL.
 */
namespace Vocal 
{


/** Infrastructure in VOCAL related to "algorithms and the assignment of
 *  responsibilities between objects".<br><br>
 *
 *  See Design Patterns, ISBN 0-201-63361-2, Chapter 5.
 */
namespace Behavioral
{


class Stimulus;
class StateInterface;


/** The basis of a finite state machine.<br><br>
 *
 *  The context is responsible for maintaining the current state of the
 *  finite state machine, and provide and means to set and verify the
 *  starting and stopping states of that machine. Users of finite
 *  state machines will typically only interface with the finite state
 *  machine class that extends this class.
 *
 *  @see    Vocal::Behavioral::StateInterface
 *  @see    Vocal::Behavioral::Stimulus
 */
class Context : public Vocal::IO::Writer
{

    protected:

    	/** Default constructor
	 */
    	Context();

    public:


    	/** Virtual destructor
	 */
	virtual     ~Context();


    	/** Reset the current state back to the starting state.
	 */
    	void	    	        reset() throw();


    	/** Reset the current state to the given state.
	 */
    	void	    	    	reset(StateInterface &) throw();


    	/** Set the starting state to the given state.
	 */
	void	    	    	startState(StateInterface &);


    	/** Query the context whether the current state is the starting state.
	 */
	bool	    	    	inStartState() const;


    	/** Set the stopping state to the given state.
	 */
	void	    	    	stopState(StateInterface &);


    	/** Query the context whether the current state is the stopping state.
	 */
	bool	    	    	inStopState() const;


    	/** Pass a stimulus to the current state.
	 */
    	void        	    	stimulate(const Stimulus &) throw ();


    	/** Write the context to am ostream.
	 */
	virtual ostream &       writeTo(ostream &) const;


    	/** The name of this class of context.
	 */
    	virtual const char *	name() const = 0;


    	/** Query the context whether the current state is the same as the 
	 *  given state.
	 */
    	bool	    	verifyState(const StateInterface &);


    private:
    
    	StateInterface	    *	myState;
	StateInterface	    *	myStartState;
	StateInterface	    *	myStopState;
};


} // namespace Behavioral
} // namespace Vocal


#endif // !defined(VOCAL_CONTEXT_HXX)
