#if !defined(VOCAL_OBSERVER_HXX)
#define VOCAL_OBSERVER_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const Observer_hxx_Version = 
    "$Id: Observer.hxx,v 1.2 2001/08/10 04:02:04 icahoon Exp $";


#include "Writer.hxx"
#include "NonCopyable.hxx"


/** Infrastructure common to VOCAL.
 */
namespace Vocal 
{


/** Infrastructure in VOCAL related to "algorithms and the assignment of
 *  responsibilities between objects".<br><br>
 *
 *  See Design Patterns, ISBN 0-201-63361-2, Chapter 5.
 */
namespace Behavioral
{


class Subject;


/** Defines a relationship between an Observer and a Subject such that when
 *  the Subject changes, the Observer will be "notified and updated 
 *  automatically".<br><br> 
 *
 *  See Design Patterns, Gamma et al, page 293. This is meant as an 
 *  interface class only. The user should extend this class as needed.<br><br>
 *
 *  For simplicity, Observers are not copyable, and each instance
 *  is considered unique.
 *
 *  @see    Vocal::Behavioral::Subject
 */
class Observer : public Vocal::IO::Writer, public Vocal::NonCopyable
{
    public:
    
    	Observer();
	virtual ~Observer();
	

    	/** The user will extend the interface using update().
	 *  When attached to a Subject and the Subject changes state,
	 *  this method will be called by the Subject, allowing the
	 *  Observer to synchronize with the Subject.
	 */
	virtual void	    update(Subject &);
	

    	/** Since each Observer is unique the memory address of the
	 *  Observer can be used to determine it's equality.
	 */
    	bool	    	    operator==(const Observer &) const;


    	/** Since each Observer is unique the memory address of the
	 *  Observer can be used to determine it's relation to other
	 *  Observers.
	 */
    	bool	    	    operator<(const Observer &) const;
};


} // namespace Behavioral
} // namespace Vocal


#endif // !defined(VOCAL_OBSERVER_HXX)
