#if !defined(VOCAL_SUBJECT_HXX)
#define VOCAL_SUBJECT_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const Subject_hxx_Version = 
    "$Id: Subject.hxx,v 1.2 2001/08/10 04:02:04 icahoon Exp $";


#include "Writer.hxx"
#include <vector>


/** Infrastructure common to VOCAL.
 */
namespace Vocal 
{


/** Infrastructure in VOCAL related to "algorithms and the assignment of
 *  responsibilities between objects".<br><br>
 *
 *  See Design Patterns, ISBN 0-201-63361-2, Chapter 5.
 */
namespace Behavioral
{


using std::vector;


class Observer;


/** Defines a relationship between a Subject and an Observer such that when
 *  the Subject changes, the Observer will be "notified and updated 
 *  automatically".<br><br>
 *
 *  See Design Patterns, Gamma et al, page 293.
 *
 *  @see    Vocal::Behavioral::Observer
 */
class Subject : public Vocal::IO::Writer
{
    public:
    
    	/** Default constructor.
	 */
    	Subject();


    	/** Virtual destructor.
	 */
	virtual ~Subject();


    	/** Attach an observer to the subject. When the subject changes
	 *  the observer will be updated. The same observer can be
	 *  added multiple times, and will be updated multiple times.
	 */	
    	void	    	    attach(Observer &);


    	/** Detach an observer from the subject. If the observer was
	 *  attached multiple times, the first one will be removed.
	 */
	void	    	    detach(Observer &);


    	/** Notify all observer's of a change.
	 */
    	void	    	    notify();


    	/** Equality relational operator. The address of the object is
	 *  used for comparison, since each Subject is considered unique.
	 */
	bool	    	    operator==(const Subject &) const;


    	/** Write the Subject to an ostream.
	 */
	virtual ostream &   writeTo(ostream &) const;


    private:

    	vector<Observer *>  myObservers;
};


} // namespace Behavioral
} // namespace Vocal


#endif // !defined(VOCAL_SUBJECT_HXX)
