#if !defined(VOCAL_CONTROL_EVENT_HXX)
#define VOCAL_CONTROL_EVENT_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const ControlEvent_hxx_Version =
    "$Id: ControlEvent.hxx,v 1.5 2001/11/01 15:57:10 sprajpat Exp $";


#include "Event.hxx"
#include "Sptr.hxx"
#include "FifoBase.h"
#include "Mutex.hxx"


/** Infrastructure common to VOCAL.
 */
namespace Vocal 
{


/** Infrastructure common to VOCAL to provide event based services.<br><br>
 */
namespace Services
{


using Vocal::Threads::Mutex;


/** ControlEvent is used to manage the starting, stopping and shutting down of
 *  a thread that waits on Fifo<Event>'s.<br><br>
 *
 *  It is best to only use these within an Sptr.
 *
 *  @see    Vocal::Services::Event
 *  @see    FifoBase
 */
class ControlEvent : public Vocal::Services::Event
{
    public:


        /** The type of control event.
         */
        enum Type
        {
            START = 0,
            STOP,
            SHUTDOWN,
            START_RESPONSE,
            STOP_RESPONSE,
            SHUTDOWN_RESPONSE,

            TOTAL_TYPES,
            
            MAX_REQUEST = SHUTDOWN,
            MIN_RESPONSE = START_RESPONSE,
            MAX_RESPONSE = SHUTDOWN_RESPONSE
    	};


        /** Create a control event with a type and an optional reply
         *  fifo. If the reply fifo is set, the control event can be
         *  used as a request (rather than a command), and the response
         *  can be communicated via the returncode.
         */
        ControlEvent(Type, FifoBase< Sptr<Event> > * event = 0);


        /**  Create a control event with a type and a reply
         *  fifo. The control event is used as a request (rather than a 
	 *  command), and the response can be communicated via the returncode.
         */
        ControlEvent(Type, FifoBase< Sptr<Event> > &);


    	/** Virtual destructor.
	 */
        virtual ~ControlEvent();


    	/** Name of the this class of event, being "ControlEvent".
	 */
        virtual const char* const name() const;


    	/** Return the type of ControlEvent.
	 */
        Type	getType() const;


    	/** Set the return code for a returned response.
	 */
        void	setReturnCode(int);


    	/** Get the return code for a returned response.
	 */
        int     getReturnCode() const;


        /** Is a request event.
         */
        bool    isRequest() const;
        

        /** Is a response event.
         */
        bool    isResponse() const;


        /** Convert the request to a response
         */
        void    response();
        
        
    	/** Set the reply fifo for a returned response.
	 */
        void	setReplyFifo(
            FifoBase< Sptr<Event> > &
        );


    	/** Get the return code for a returned response.
	 */
        FifoBase< Sptr<Event> > * getReplyFifo() const;


        /** To uniquely identify a request. Useful if this control event is
         *  a request. The fifo waiting for the response can the uniquely
         *  identify the response, in the case they sent out multiple control
         *  events simultaneously.
         */
        int     getControlEventId() const;


        /** Write the event to a stream.
         */
        virtual ostream & writeTo(ostream &) const;


    private:

        // nextId is a globally shared resource, so it needs to be protected.
        //
        static	Mutex	                myNextIdMutex;
        static	int                     myNextId;

        Type	                        myType;
        int                             myReturnCode;
        FifoBase < Sptr < Event > > *	myFifo;
        int                             myId;
};


} // namespace Services
} // namespace Vocal


#endif // !defined(VOCAL_CONTROL_EVENT_HXX)
