#if !defined(VOCAL_SERVICE_APPLICATION_HXX)
#define VOCAL_SERVICE_APPLICATION_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const ServiceApplication_hxx_Version = 
    "$Id: ServiceApplication.hxx,v 1.5 2002/03/16 04:55:32 icahoon Exp $";


#include "Application.hxx"
#include "Sptr.hxx"
#include "FifoBase.h"
#include <vector>
#include <string>


namespace Vocal {
    namespace Signals {
        class SignalSet;
        class SignalAction;
    }
    namespace Configuration {
        class Config;
    }
}
    


/** Vovida Open Communication Application Library.<br><br>
 */
namespace Vocal 
{


/** Infrastructure common to VOCAL to provide event based services.<br><br>
 */
namespace Services
{


using Vocal::Signals::SignalSet;
using Vocal::Signals::SignalAction;
using Vocal::Configuration::Config;


class ServiceManager;
class Event;


/** Service based application.<br><br>
 */
class ServiceApplication : public Vocal::Application
{
    public:

    	/** Create the service application. Initialization is done in init, 
         *  not here.
	 */
    	ServiceApplication(bool singleThreaded = false);


    	/** Destroy the service application. Uninitialization is done in 
         *  uninit, not here.
    	 */
    	virtual ~ServiceApplication();


    	/** Initialize the service application.
    	 */
    	ReturnCode  	    init(int, char **, char **);

    	
	/** Cleanup.
    	 */
	void	    	    uninit();
	
    	
	/** Run the service application.
    	 */
    	ReturnCode  	    run();


    protected:


        /** The user should provide a factory method for the user defined
         *  configuration. If the user doesn't provide this, the default
         *  (and somewhat incomplete) Config class will be used.
         */
        virtual Config *    createConfig();
        
        
        /** The user has the ability to execute code after the configuration
         *  has been loaded and before initialization begins.
         */
        virtual ReturnCode  postConfig();
        
        
        /** The user can optionally provide the set of signals that will 
         *  cause the application to terminate gracefully. The default 
         *  signal set will include SIGHUP, SIGINT and SIGTERM.
         */
        virtual const vector<int> &     terminateSignals() const;
        
        
        /** The user can provide the filename of the log. 
         *  The config has been created when this is called.
         */
        virtual string  logFile() const;
         
        
        /** The last thing called by init to give the user a chance to
         *  initialize. All server objects have been initialized when
         *  this is called.
         */
        virtual ReturnCode  userInit();


        /** The first thing called by uninit to give the user a chance to
         *  cleanup.
         */
        virtual void    userUninit();
        
        
        /** Config accessor
         */        
        Config &    config();
        
        
        /** Config accessor
         */
        const Config &  config() const;


        /** Signal action accessor
         */
        SignalAction &  signalAction();
          
  
        /** Fifo accessor
         */ 
        FifoBase< Sptr<Event> > &   fifo();
        
        
        /** ServiceManager accessor
         */
        ServiceManager &    serviceManager();      


    private:
    
    	/** Server configuration information.
	 */
	Config              *   myConfig;
	
	
    	/** The set of signals we are observing.
	 */
	SignalSet   	    *	myTerminateSignals;


    	/** The signal handler for the pervious signal set.
	 */
	SignalAction     *	myTerminateAction;


    	/** Return message fifo. Used by the service manager to 
	 *  communicate with the subordinate threads / services.
	 */
	FifoBase< Sptr<Event> > *   myFifo;


    	/** Mananges all services. It starts, stops and shutsdown
	 *  the services.
	 */
    	ServiceManager  	*   myServiceMgr;
        

        /** If set, the application is running single threaded.
         */
        bool                        mySingleThreaded;
};


} // namespace Services
} // namespace Vocal


#endif // !defined(VOCAL_SERVICE_APPLICATION_HXX)
