/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCA;L", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const Condition_cxx_Version =
    "$Id: Condition.cxx,v 1.2 2001/08/10 04:02:08 icahoon Exp $";


#include "global.h"
#include "Condition.hxx"
#include "Mutex.hxx"
#include "TimeVal.hxx"
#include "VocalCommon.hxx"
#include <cassert>


using Vocal::Threads::Condition;
using Vocal::Threads::Mutex;
using Vocal::TimeAndDate::TimeVal;
using Vocal::ReturnCode;
using Vocal::SUCCESS;


Condition::Condition()
{
    ReturnCode  rc = vcond_init(&myId);
    assert( rc == SUCCESS );
}


Condition::~Condition ()
{
    ReturnCode  rc = vcond_destroy(&myId);
    assert( rc == SUCCESS );
}


int
Condition::wait (Mutex* mutex, int relativeTimeInMs)
{
    int utime = relativeTimeInMs < 1
                    ? relativeTimeInMs
                    : relativeTimeInMs * 1000;
    
    return ( uwait(mutex, utime) );
}


int
Condition::uwait (Mutex* mutex, int relativeTimeInUs)
{
    // SunOS 5.X (according to the man page for pthread_cond_timedwait)
    // can have a maximum value for the seconds portion of the abstime
    // of the current time + 100,000,000. The nsec can have a maximum
    // value of 1,000,000,000. You get EINVAL otherwise.
    //
    // Note that the relative time is in milliseconds, meaning
    // that even if we use MAXINT (assuming 32 bit ints), then
    // we get a number well below MAXINT (by a couple orders of
    // magnitude) for the seconds portion. The slightly challenging
    // part is making sure the nsec is less than 1,000,000,000.
    //
    if ( relativeTimeInUs < 0 )
    {
        return ( vcond_wait(&myId, mutex->getId()) );
    }

    // Relative time is specified, so create the absolute time of expiration.
    //
    TimeVal expiresTV;
    expiresTV.now();
    
    timeval     relativeTime;
    relativeTime.tv_sec = 0;
    relativeTime.tv_usec = relativeTimeInUs;

    expiresTV += relativeTime;

    timespec	expiresTS;

    expiresTS.tv_sec = expiresTV.tv_sec;
    expiresTS.tv_nsec = expiresTV.tv_usec * 1000;

    assert( expiresTV.tv_usec < 1000000000L );

    return ( vcond_timedwait(&myId, mutex->getId(), &expiresTS) );
}


int
Condition::signal ()
{
    return ( vcond_signal(&myId) );
}


int
Condition::broadcast()
{
    return ( vcond_broadcast(&myId) );
}


const vcondition_t*
Condition::getId () const
{
    return ( &myId );
}


void vusleep(unsigned long usec)
{
    Mutex      mutex;
    Condition  cond;
    
    mutex.lock();
    cond.uwait(&mutex, usec);
    mutex.unlock();
}
